#include <config.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "pgc.h"
#include "pgc_function.h"
#include "pgc_function_interface.h"
#include "pgc_hash.h"

struct pgc_function_definition **function_table = NULL;

/* An internal function, used purely by init_function_table () and the people
   who love it. */
void pgc_add_function (char *function_name, void *function_pointer, 
		       int return_type, int arg_count, ...) {
  va_list arg_types;
  int i;
  int hashed_index = pgc_hash (function_name, PGC_FUNCTION_TABLE_SIZE);
  int insertion_index = -1;
  for (i = hashed_index; i < PGC_FUNCTION_TABLE_SIZE; i++) {
    if (function_table[i] == NULL) {
      insertion_index = i;
      break;
    }
    /* Functions should not be getting redefined... */
    if (strcmp (function_name, function_table[i]->function_name) == 0) 
      return;
  }
  if (insertion_index == -1)
    for (i = hashed_index; i < PGC_FUNCTION_TABLE_SIZE; i++) {
      if (function_table[i] == NULL) {
	insertion_index = i;
	break;
      }
      /* As above... */
      if (strcmp (function_name, function_table[i]->function_name) == 0)
	return;
    }
  if (insertion_index != -1) {
    function_table[insertion_index] = 
      malloc (sizeof (struct pgc_function_definition));
    function_table[insertion_index]->function_name = strdup (function_name);
    function_table[insertion_index]->function_pointer = function_pointer;
    function_table[insertion_index]->return_type = return_type;
    function_table[insertion_index]->arg_count = arg_count;
    if (arg_count == 0)
      function_table[insertion_index]->arg_types = NULL;
    else {
      int j;
      function_table[insertion_index]->arg_types = 
	malloc (sizeof (int) * arg_count);
      va_start (arg_types, arg_count);
      for (j = 0; j < arg_count; j++)
	function_table[insertion_index]->arg_types[j] = 
	  va_arg (arg_types, int);
      va_end (arg_types);
    }
  }
  return;
}

struct pgc_function_definition *pgc_lookup_function (char *function_name) {
  int i;
  int hashed_index = pgc_hash (function_name, PGC_FUNCTION_TABLE_SIZE);
  for (i = hashed_index; i < PGC_FUNCTION_TABLE_SIZE; i++) {
    if (function_table[i] == NULL)
      return NULL;
    if (strcmp (function_name, function_table[i]->function_name) == 0)
      return function_table[i];
  }
  for (i = 0; i < hashed_index; i++) {
    if (function_table[i] == NULL)
      return NULL;
    if (strcmp (function_name, function_table[i]->function_name) == 0)
      return function_table[i];
  }
  return NULL;
}

void pgc_init_function_table () {
  function_table = 
    calloc (PGC_FUNCTION_TABLE_SIZE,
	    sizeof (struct pgc_function_definition *)); 
  pgc_add_function ("version", pgc_internal_version, PGC_STRING, 0);
  pgc_add_function ("print", pgc_internal_print, PGC_NIL, 1, PGC_STRING);
  pgc_add_function ("int", pgc_internal_int, PGC_INT, 1, PGC_FLOAT);
  pgc_add_function ("float", pgc_internal_float, PGC_FLOAT, 1, PGC_INT);
  pgc_add_function ("int_to_string", pgc_internal_int_to_string, PGC_STRING, 1,
		    PGC_INT);
  pgc_add_function ("float_to_string", pgc_internal_float_to_string, 
		    PGC_STRING, 1, PGC_FLOAT);
  pgc_add_function ("record_type", pgc_internal_record_type, PGC_STRING, 1,
		    PGC_RECORD);
  pgc_add_function ("describe", pgc_internal_describe, PGC_NIL, 1, PGC_STRING);
  pgc_add_function ("vector_sum", pgc_internal_vector_sum, PGC_FLOAT, 1,
		    PGC_VECTOR);
  pgc_add_function ("vector_type", pgc_internal_vector_type, PGC_STRING, 1,
		    PGC_VECTOR);
  pgc_add_function ("vector_length", pgc_internal_vector_length, PGC_INT, 1,
		    PGC_VECTOR);

  #ifdef HAVE_LIBPROTEINGEOMETRY
  pgc_add_function ("read_pdb", pgc_interface_volume_read_pdb, PGC_RECORD, 1,
		    PGC_STRING);
  pgc_add_function ("volume", pgc_interface_volume_volume, PGC_VECTOR, 1,
		    PGC_RECORD);
  pgc_add_function ("surface", pgc_interface_volume_surface, PGC_VECTOR, 1,
		    PGC_RECORD);
  pgc_add_function ("packing_eff", pgc_interface_volume_packing_eff, 
		    PGC_VECTOR, 1, PGC_RECORD);
  pgc_add_function ("packing_eff_advanced", 
		    pgc_interface_volume_packing_eff_advanced, PGC_VECTOR, 4,
		    PGC_RECORD, PGC_STRING, PGC_STRING, PGC_STRING);
  #endif
  
  #ifdef HAVE_LSQRMS
  pgc_add_function ("lsqrms", pgc_interface_lsqrms, PGC_FLOAT, 2, PGC_STRING,
		    PGC_STRING);
  #endif
}

