#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pgc.h"
#include "pgc_function_interface.h"
#include "pgc_record.h"

/* The caller pretty much has to guarantee that the correct number of arguments
   will be returned... */

/* Here are the internal PGC functions */
void *pgc_internal_version (void **args) {
  return strdup (PACKAGE_VERSION);
}

void *pgc_internal_print (void **args) {
  printf ("%s\n", (char *) args[0]);
  return NULL;
}

void *pgc_internal_int (void **args) {
  int *return_value = malloc (sizeof (int));
  *return_value = (int) *((double *) args[0]);
  return return_value;
}

void *pgc_internal_float (void **args) {
  double *return_value = malloc (sizeof (double));
  *return_value = (double) *((int *) args[0]);
  return return_value;
}

void *pgc_internal_int_to_string (void **args) {
  /* Should we use a #define for this value? */
  char *return_string = malloc (sizeof (char *) * 64);
  (void) sprintf (return_string, "%d", *((int *) args[0]));
  return_string = 
    realloc (return_string, sizeof (char *) * (strlen (return_string) + 1));
  return return_string;
}

void *pgc_internal_float_to_string (void **args) {
  char *return_string = malloc (sizeof (char *) * 64);
  (void) sprintf (return_string, "%lf", *((double *) args[0]));
  return_string = 
    realloc (return_string, sizeof (char *) * (strlen (return_string) + 1));
  return return_string;
}

void *pgc_internal_describe (void **args) {
  void *scratch;
  int i;
  if ((args != NULL) && (args[0] != NULL)) {
    scratch = pgc_lookup_record ((char *) args[0]);
    if (scratch != NULL) {
      if (((struct pgc_record_definition *) scratch)->field_count > 0) {
	printf 
	  ("describe: record type \"%s\" has the following public fields:\n",
		(char *) args[0]);
	for (i = 0; 
	     i < ((struct pgc_record_definition *) scratch)->field_count;
	     i++)
	  printf ("  %s\n", ((struct pgc_record_definition *) scratch)->
		  public_field_names[i]);
      }
      else
	printf ("describe: record type \"%s\" has no public fields.\n", 
		(char *) args[0]);
    }
    else
      printf ("describe: no definition found for record type \"%s\".\n", 
	      (char *) args[0]);
  }
  return NULL;
}

void *pgc_internal_vector_sum (void **args) {
  int i;
  double *result = malloc (sizeof (double));
  *result = 0.0;
  if ((args != NULL) && (args[0] != NULL)) {
    /* So are we going to check to see whether args[0] is a number vector? */
    switch (((struct pgc_vector *) args[0])->vector_type) {
    case PGC_INT:
      for (i = 0; i < ((struct pgc_vector *) args[0])->vector_length; i++)
	*result += *(((int **) ((struct pgc_vector *) args[0])->vector)[i]);
      break;
    case PGC_FLOAT:
      for (i = 0; i < ((struct pgc_vector *) args[0])->vector_length; i++)
	*result += *(((double **) ((struct pgc_vector *) args[0])->vector)[i]);
      break;
    default:
      fprintf (stderr, "vector_sum: vector must be of type int or float.\n");
      *result = strtod ("NaN", NULL);
      break;
    } 
  }
  return (void *) result;
}

void *pgc_internal_vector_type (void **args) {
  switch (((struct pgc_vector *) args[0])->vector_type) {
  case PGC_NIL:
    return (strdup ("nil"));
  case PGC_INT:
    return (strdup ("int"));
  case PGC_FLOAT:
    return (strdup ("float"));
  case PGC_STRING:
    return (strdup ("string"));
  case PGC_RECORD:
    return (strdup ("record"));
  case PGC_VECTOR:
    return (strdup ("vector"));
  default:
    return strdup ("");
  }
}

void *pgc_internal_record_type (void **args) {
  if ((((struct pgc_record *) args[0])->record_type_name != NULL) &&
      (strcmp (((struct pgc_record *) args[0])->record_type_name, "") != 0))
    return strdup (((struct pgc_record *) args[0])->record_type_name);
  else 
    return strdup ("null");
}

void *pgc_internal_vector_length (void **args) {
  int *result = malloc (sizeof (int));
  *result = ((struct pgc_vector *) args[0])->vector_length;
  return result;
}
